Package: rsurveycto
Type: Package
Title: Interact with Data on 'SurveyCTO'
Version: 0.2.6
Authors@R: c(
  person('Jake', 'Hughey', , 'jake@agency.fund', c('aut', 'cre')),
  person('Robert', 'On', , 'robert@agency.fund', 'aut'))
Description: 'SurveyCTO' is a platform for mobile data collection in offline settings.
  The 'rsurveycto' R package uses the 'SurveyCTO' REST API
  <https://docs.surveycto.com/05-exporting-and-publishing-data/05-api-access/01.api-access.html>
  to read datasets and forms from a 'SurveyCTO' server into R as 'data.table's
  and to download file attachments. The package also has limited support to
  write datasets to a server.
URL: https://agency-fund.github.io/rsurveycto/,
        https://github.com/agency-fund/rsurveycto
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1)
Imports: checkmate (>= 2.3.2), cli (>= 3.6.3), curl (>= 5.2.1),
        data.table (>= 1.15.4), glue (>= 1.7.0), httr (>= 1.4.7),
        jsonlite (>= 1.8.8), lifecycle (>= 1.0.4), readxl (>= 1.4.3),
        rlang (>= 1.1.4), vctrs (>= 0.6.5), withr (>= 3.0.1)
Suggests: knitr, rmarkdown, testthat (>= 3.2.1.1)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-09 11:26:40 UTC; jakehughey
Author: Jake Hughey [aut, cre],
  Robert On [aut]
Maintainer: Jake Hughey <jake@agency.fund>
Repository: CRAN
Date/Publication: 2026-04-09 11:40:02 UTC
Built: R 4.6.0; ; 2026-04-29 07:29:33 UTC; unix

Package: rsurveycto
Type: Package
Title: Interact with Data on 'SurveyCTO'
Version: 0.2.6
Authors@R: c(
  person('Jake', 'Hughey', , 'jake@agency.fund', c('aut', 'cre')),
  person('Robert', 'On', , 'robert@agency.fund', 'aut'))
Description: 'SurveyCTO' is a platform for mobile data collection in offline settings.
  The 'rsurveycto' R package uses the 'SurveyCTO' REST API
  <https://docs.surveycto.com/05-exporting-and-publishing-data/05-api-access/01.api-access.html>
  to read datasets and forms from a 'SurveyCTO' server into R as 'data.table's
  and to download file attachments. The package also has limited support to
  write datasets to a server.
URL: https://agency-fund.github.io/rsurveycto/,
        https://github.com/agency-fund/rsurveycto
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1)
Imports: checkmate (>= 2.3.2), cli (>= 3.6.3), curl (>= 5.2.1),
        data.table (>= 1.15.4), glue (>= 1.7.0), httr (>= 1.4.7),
        jsonlite (>= 1.8.8), lifecycle (>= 1.0.4), readxl (>= 1.4.3),
        rlang (>= 1.1.4), vctrs (>= 0.6.5), withr (>= 3.0.1)
Suggests: knitr, rmarkdown, testthat (>= 3.2.1.1)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-09 11:26:40 UTC; jakehughey
Author: Jake Hughey [aut, cre],
  Robert On [aut]
Maintainer: Jake Hughey <jake@agency.fund>
Repository: CRAN
Date/Publication: 2026-04-09 11:40:02 UTC
Built: R 4.5.2; ; 2026-04-15 05:09:35 UTC; unix

# Generated by roxygen2: do not edit by hand

export(drop_empties)
export(scto_auth)
export(scto_catalog)
export(scto_get_attachments)
export(scto_get_form_definitions)
export(scto_get_form_metadata)
export(scto_get_form_responses)
export(scto_meta)
export(scto_quiet)
export(scto_read)
export(scto_unnest_form_definitions)
export(scto_write)
import(checkmate)
import(cli)
import(data.table)
importFrom(glue,glue)
importFrom(httr,GET)
importFrom(httr,POST)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,cookies)
importFrom(httr,headers)
importFrom(httr,set_cookies)
importFrom(lifecycle,deprecated)

get_session_auth = function(servername, username, password) {
  index_url = glue('https://{servername}.surveycto.com/index.html')
  index_res = GET(index_url)
  csrf_token = headers(index_res)$`x-csrf-token`

  if (is.null(csrf_token)) {
    scto_abort(paste(
      'Unable to access server {.server {servername}}.',
      'Please check that server is running.'))
  }

  login_url = glue(
    'https://{servername}.surveycto.com/login?spring-security-redirect=%2F')
  login_res = POST(
    login_url,
    body = list(
      username = username,
      password = password,
      csrf_token = csrf_token),
    encode = 'form')

  scto_cookies = cookies(login_res)
  session_id = scto_cookies$value[scto_cookies$name == 'JSESSIONID']
  session_auth = list(csrf_token = csrf_token, session_id = session_id)
  session_auth
}


#' Authenticate with a SurveyCTO server
#'
#' SurveyCTO's API supports basic authentication using a username and password.
#' Make sure the user is assigned a role with permission to download data
#' ("data manager" or greater) and "Allow server API access" is enabled.
#'
#' @param auth_file String indicating path to a text file containing the
#'   server name on the first line, username on the second, and password on the
#'   third. Other arguments are only used if `auth_file` is `NULL`.
#' @param servername String indicating name of the SurveyCTO server.
#' @param username String indicating username for the SurveyCTO account.
#' @param password String indicating password for the SurveyCTO account.
#' @param validate Logical indicating whether to validate credentials by calling
#'   [scto_meta()]. Should only be set to `FALSE` for debugging.
#'
#' @return `scto_auth` object for an authenticated SurveyCTO session.
#'
#' @examples
#' \dontrun{
#' # preferred approach
#' auth = scto_auth('scto_auth.txt')
#'
#' # alternate approach
#' auth = scto_auth('my_server', 'my_user', 'my_pw', auth_file = NULL)
#' }
#'
#' @export
scto_auth = function(
    auth_file = NULL, servername = NULL, username = NULL, password = NULL,
    validate = TRUE) {

  if (is.null(auth_file)) {
    assert_string(servername)
    assert_string(username)
    assert_string(password)
  } else {
    assert_string(auth_file)
    assert_file_exists(auth_file)
    auth_char = readLines(auth_file, warn = FALSE)
    if (!test_character(auth_char, any.missing = FALSE, len = 3L)) {
      scto_abort(paste(
        '`auth_file` {.filename {auth_file}} must have exactly three lines:',
        'servername, username, and password.'))
    }
    servername = auth_char[1L]
    username = auth_char[2L]
    password = auth_char[3L]
  }

  assert_flag(validate)

  handle = curl::new_handle()
  curl::handle_setopt(
    handle = handle,
    httpauth = 1,
    userpwd = glue('{username}:{password}'))

  session_auth = get_session_auth(servername, username, password)

  auth = list(servername = servername, handle = handle,
              csrf_token = session_auth$csrf_token,
              session_id = session_auth$session_id)
  class(auth) = 'scto_auth'

  if (validate) invisible(scto_meta(auth))
  auth
}

if (identical(Sys.getenv('NOT_CRAN'), 'true')) { # !on_cran()
  auth = scto_auth(auth_file = auth_file)
  d = scto_read(auth, 'hh_listing_example_1')
  d = d[(.N - 2):.N]
  d_enc = scto_read(
    auth, 'hh_example_encrypted', private_key = private_key)
}

test_that('scto_get_attachments ok', {
  skip_on_cran()
  filenames = scto_get_attachments(
    auth, d$household_photo, output_dir = output_dir)
  expect_file_exists(file.path(output_dir, filenames[!is.na(filenames)]))
  expect_error(scto_get_attachments(
    auth, d$household_photo, output_dir = output_dir, overwrite = FALSE))
})

test_that('scto_get_attachments empty', {
  skip_on_cran()
  filenames = scto_get_attachments(auth, d$district, output_dir = output_dir)
  expect_true(allMissing(filenames))
})

test_that('scto_get_attachments encrypted key', {
  skip_on_cran()
  filenames = scto_get_attachments(
    auth, d_enc$household_photo, output_dir = output_dir,
    private_key = private_key)
  expect_file_exists(file.path(output_dir, filenames[!is.na(filenames)]))
  expect_snapshot_file(file.path(output_dir, filenames[1L]))
})

test_that('scto_get_attachments encrypted no key', {
  skip_on_cran()
  filenames = scto_get_attachments(
    auth, d_enc$household_photo, output_dir = output_dir)
  expect_file_exists(file.path(output_dir, filenames[!is.na(filenames)]))
  expect_snapshot_file(file.path(output_dir, filenames[2L]))
})

test_that('get_session_auth ok', {
  skip_on_cran()
  sess = get_session_auth(auth_args[1L], auth_args[2L], auth_args[3L])
  expect_type(sess, 'list')
})

test_that('get_session_auth not ok', {
  skip_on_cran()
  expect_error(get_session_auth('1', 'nobody@a.io', 'password'))
})

test_that('scto_auth file ok', {
  skip_on_cran()
  auth = scto_auth(auth_file = auth_file)
  expect_class(auth, 'scto_auth')
})

test_that('scto_auth file not ok', {
  skip_on_cran()
  f = withr::local_tempfile()
  writeLines(c('swans', 'elephants'), f)
  expect_error(scto_auth(auth_file = f))
})

test_that('scto_auth args', {
  skip_on_cran()
  auth = scto_auth(
    servername = auth_args[1L], username = auth_args[2L],
    password = auth_args[3L])
  expect_class(auth, 'scto_auth')
})

if (identical(Sys.getenv('NOT_CRAN'), 'true')) { # !on_cran()
  auth = scto_auth(auth_file = auth_file)
}

test_that('scto_get_responses ok', {
  skip_on_cran()
  d = scto_get_form_responses(auth, 'hh_listing_example_1')
  expect_data_table(d)
})

if (identical(Sys.getenv('NOT_CRAN'), 'true')) { # !on_cran()
  auth = scto_auth(auth_file = auth_file)
  nms = c('settings', 'choices', 'fields', 'formulasConvertedToStaticValues')
}

test_that('scto_get_form_definitions one simplify', {
  skip_on_cran()
  def = scto_get_form_definitions(
    auth, 'hh_listing_example_1', simplify = TRUE)
  expect_type(def, 'list')
  expect_named(def, nms)
})

test_that('scto_get_form_definitions one not simplify', {
  skip_on_cran()
  defs = scto_get_form_definitions(
    auth, 'hh_listing_example_1', simplify = FALSE)
  expect_type(defs, 'list')
  expect_named(defs, 'hh_listing_example_1')
})

test_that('scto_get_form_definitions all', {
  skip_on_cran()
  defs = scto_get_form_definitions(auth)
  expect_type(defs, 'list')
  expect_named(defs)
})

test_that('scto_get_form_definitions not ok', {
  skip_on_cran()
  expect_error(scto_get_form_definitions(auth, 'flux_capacitors'))
})

if (identical(Sys.getenv('NOT_CRAN'), 'true')) { # !on_cran()
  auth = scto_auth(auth_file = auth_file)
}

test_that('scto_read dataset ok', {
  skip_on_cran()
  d = scto_read(auth, 'enumerators')
  expect_data_table(d)
})

test_that('scto_read dataset drop empties', {
  skip_on_cran()
  d = scto_read(auth, 'enumerators', drop_empty_cols = TRUE)
  expect_data_table(d)
})

test_that('scto_read dataset not ok', {
  skip_on_cran()
  expect_error(scto_read(auth, 'flux_capacitors'))
})

test_that('scto_read form ok', {
  skip_on_cran()
  d = scto_read(auth, 'hh_listing_example_1')
  expect_data_table(d)
})

test_that('scto_read form not ok', {
  skip_on_cran()
  expect_error(scto_read(auth, 'hh_listing_example_1', start_date = 'coffee'))
})

test_that('scto_read form encrypted no key', {
  skip_on_cran()
  d = scto_read(auth, 'hh_example_encrypted')
  expect_data_table(d, ncols = 10L)
})

test_that('scto_read form encrypted key', {
  skip_on_cran()
  d = scto_read(auth, 'hh_example_encrypted', private_key = private_key)
  expect_data_table(d, ncols = 21L)
})

test_that('scto_read not simplify', {
  skip_on_cran()
  db = scto_read(auth, 'enumerators', simplify = FALSE)
  expect_list(db, min.len = 1L)
  lapply(db, function(d) expect_data_table(d))
})

test_that('scto_read all', {
  skip_on_cran()
  db = scto_read(auth)
  expect_list(db, min.len = 1L)
})

if (identical(Sys.getenv('NOT_CRAN'), 'true')) { # !on_cran()
  auth = scto_auth(auth_file = auth_file)
  dataset_id = 'wallahs'
  d0 = scto_read(auth, dataset_id)
}

set.seed(as.integer(Sys.time()))

test_that('scto_write no exist', {
  skip_on_cran()
  skip_if(Sys.getenv('GITHUB_ACTIONS') == 'true') # avoid concurrent changes
  expect_error(scto_write(auth, 'rhinos'))
})

test_that('scto_write no append', {
  skip_on_cran()
  skip_if(Sys.getenv('GITHUB_ACTIONS') == 'true') # avoid concurrent changes
  withr::defer(scto_write(auth, d0, dataset_id, append = FALSE))

  d1 = copy(d0)
  d1[, mass := round(runif(.N), 5)]
  r = scto_write(auth, d1, dataset_id, append = FALSE)
  d2 = scto_read(auth, dataset_id)

  expect_named(r, c('data_old', 'response'))
  expect_identical(d0, r$data_old)
  expect_identical(d1, d2)
  expect_s3_class(r$response, 'response')
})

test_that('scto_write no fill', {
  skip_on_cran()
  skip_if(Sys.getenv('GITHUB_ACTIONS') == 'true') # avoid concurrent changes
  d1 = copy(d0)
  d1[, tree := NULL]
  expect_error(scto_write(auth, d1, dataset_id, append = TRUE, fill = FALSE))
})

test_that('scto_write fill', {
  skip_on_cran()
  skip_if(Sys.getenv('GITHUB_ACTIONS') == 'true') # avoid concurrent changes
  withr::defer(scto_write(auth, d0, dataset_id, append = FALSE))

  d1 = copy(d0)
  d1[, tree := NULL]
  r = scto_write(auth, d1, dataset_id, append = TRUE, fill = TRUE)
  d2 = scto_read(auth, dataset_id)

  d3 = rbind(d0, d1, fill = TRUE)
  expect_identical(d2, d3, ignore_attr = 'scto_type')
})

#' Read data from a SurveyCTO server
#'
#' This function can read datasets and forms.
#'
#' @param auth [scto_auth()] object.
#' @param ids Character vector indicating ids of the datasets and/or forms.
#'   `NULL` indicates all datasets and forms.
#' @param start_date Date-time or something coercible to a date-time
#'   indicating the earliest date-time (UTC timezone) for which to fetch data.
#'   Only used for forms. Use with caution, because fields that are deleted
#'   prior to `start_date` will not show up, even if submissions prior to
#'   `start_date` have data for those fields.
#' @param review_status String or character vector indicating which submissions
#'   to fetch. Possible values are "approved", "pending", "rejected", or any
#'   combination of the three. Only used for forms.
#' @param private_key String indicating path to private key file. Only needs to
#'   be non-`NULL` to read encrypted form data.
#' @param drop_empty_cols Logical indicating whether to drop columns that
#'   contain only `NA` or only an empty string.
#' @param convert_datetime Character vector of column names in the data for
#'   which to convert strings to datetimes (POSIXct). Use `NULL` to not convert
#'   any columns to datetimes.
#' @param datetime_format String indicating format of datetimes from SurveyCTO.
#'   See [strptime()].
#' @param simplify Logical indicating whether to return only a `data.table`
#'   instead of a list of `data.table`s if reading one form or dataset.
#'
#' @return If `simplify` is `TRUE` and reading one form or dataset, a
#'   `data.table`. Otherwise a named list of `data.table`s, one for each form
#'   and dataset, along with a `data.table` named ".catalog" from
#'   `scto_catalog()`.
#'
#' @seealso [scto_get_form_responses()]
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' form_data = scto_read(auth, 'my_form')
#' all_data = scto_read(auth)
#' }
#'
#' @export
scto_read = function(
    auth, ids = NULL, start_date = as.POSIXct('1900-01-01', tz = 'UTC'),
    review_status = 'approved', private_key = NULL, drop_empty_cols = FALSE,
    convert_datetime = c(
      'CompletionDate', 'SubmissionDate', 'starttime', 'endtime'),
    datetime_format = '%b %e, %Y %I:%M:%S %p', simplify = TRUE) {

  assert_class(auth, 'scto_auth')
  assert_character(ids, any.missing = FALSE, unique = TRUE, null.ok = TRUE)

  start_date = as.POSIXct(start_date, tz = 'UTC')
  assert_posixct(start_date, any.missing = FALSE, len = 1L)
  start_date = max(1, as.numeric(start_date))

  review_status = match.arg(
    review_status, c('approved', 'pending', 'rejected'), several.ok = TRUE)
  review_status = paste(review_status, collapse = ',')

  assert_string(private_key, null.ok = TRUE)
  if (!is.null(private_key)) assert_file_exists(private_key)

  assert_flag(drop_empty_cols)
  assert_character(convert_datetime, any.missing = FALSE, null.ok = TRUE)
  assert_string(datetime_format)
  assert_flag(simplify)

  catalog = scto_catalog(auth)

  if (!is.null(ids) && !(all(ids %in% catalog$id))) {
    ids_bad = ids[!(ids %in% catalog$id)] # nolint
    scto_abort(paste(
      '{qty(ids_bad)} The id{?s} {.id {ids_bad}} {?was/were}',
      'not found on the server {.server {auth$servername}}.'))
  }

  catalog_now = if (is.null(ids)) catalog else catalog[catalog$id %in% ids]

  r = lapply(seq_len(nrow(catalog_now)), \(i) {
    id = catalog_now$id[i]
    if (catalog_now$type[i] == 'form') {
      scto_read_form(
        auth, id, start_date, review_status, private_key, drop_empty_cols,
        convert_datetime, datetime_format)
    } else {
      scto_read_dataset(
        auth, id, drop_empty_cols, convert_datetime, datetime_format)
    }
  })

  if (length(r) == 1L && simplify) {
    r = r[[1L]]
  } else {
    names(r) = catalog_now$id
    r$.catalog = catalog # surveycto prohibits "." in ids, so we're safe
  }
  r
}

assert_read_args = function(
    auth, drop_empty_cols, convert_datetime, datetime_format) {
  assert_class(auth, 'scto_auth')
  assert_flag(drop_empty_cols)
  assert_character(convert_datetime, any.missing = FALSE, null.ok = TRUE)
  assert_string(datetime_format)
  invisible()
}


get_resource = function(auth, type, request_url, private_key) {
  res = if (type == 'form' && !is.null(private_key)) {
    POST(request_url, body = list(private_key = httr::upload_file(private_key)))
  } else {
    curl::curl_fetch_memory(request_url, handle = auth$handle)
  }
  res
}


get_resource_retry = function(auth, type, request_url, private_key) {
  res = get_resource(auth, type, request_url, private_key)
  if (res$status_code == 409L) { # rejected due to parallel requests
    cli::cli_alert_info('Waiting for a parallel request to finish.')
    n_retry = 10
    while (n_retry > 0) {
      Sys.sleep(3)
      res = get_resource(auth, type, request_url, private_key)
      n_retry = if (res$status_code == 200L) 0 else n_retry - 1
    }
  }
  res
}


get_scto_data = function(
    auth, type, request_url, drop_empty_cols, convert_datetime, datetime_format,
    private_key = NULL) {

  res = get_resource_retry(auth, type, request_url, private_key)
  status = res$status_code
  content = rawToChar(res$content)

  if (status != 200L) {
    cli_alert_warning('Response content:\n{content}')
    scto_abort('Non-200 response: {status}')
  }

  scto_data = if (content == '') {
    data.table()
  } else if (type == 'form') {
    data.table(jsonlite::fromJSON(content, flatten = TRUE))
  } else {
    fread(text = content, na.strings = '')
  }
  setattr(scto_data, 'scto_type', type)

  if (drop_empty_cols) {
    drop_empties(scto_data)
  } else if (nrow(scto_data) > 0L) {
    idx = sapply(scto_data, \(x) is.logical(x) && all(is.na(x)))
    for (j in colnames(scto_data)[idx]) {
      set(scto_data, j = j, value = as.character(scto_data[[j]]))
    }
  }

  cols = intersect(colnames(scto_data), convert_datetime)
  for (col in cols) {
    set(scto_data, j = col, value = as.POSIXct(
      scto_data[[col]], tz = 'UTC', format = datetime_format))
  }

  scto_data[]
}


scto_read_form = function(
    auth, id, start_date, review_status, private_key, drop_empty_cols,
    convert_datetime, datetime_format) {

  request_url = glue(
    'https://{auth$servername}.surveycto.com/api/v2/forms/',
    'data/wide/json/{id}?date={start_date}&r={review_status}')

  scto_bullets(c(v = 'Reading form {.form {id}}.'))
  get_scto_data(
    auth, 'form', request_url, drop_empty_cols, convert_datetime,
    datetime_format, private_key)
}


scto_read_dataset = function(
    auth, id, drop_empty_cols, convert_datetime, datetime_format) {

  request_url = glue(
    'https://{auth$servername}.surveycto.com/api/v2/datasets/data/csv/{id}')

  scto_bullets(c(v = 'Reading dataset {.dataset {id}}.'))
  get_scto_data(
    auth, 'dataset', request_url, drop_empty_cols, convert_datetime,
    datetime_format)
}

rename_cols = \(x) gsub('[^a-zA-Z0-9]', '_', tolower(x))

#' Read form data from a SurveyCTO server and convert to long format
#'
#' This function converts data from [scto_read()] to long format, then joins
#' the data with the "survey" and "choices" components of the form metadata.
#'
#' @param auth [scto_auth()] object.
#' @param form_id String indicating id of the form.
#' @param id_cols Character vector of data columns to keep in the result, will
#'   be passed as `id.vars` to [data.table::melt()].
#' @param exclude_cols Character vector of data columns to exclude from the
#'   result.
#' @param ... Additional arguments passed to [scto_read()].
#'
#' @return A `data.table` with one row per response, i.e., one row per form
#'   submission per field. If necessary to avoid ambiguity, columns from the
#'   "survey" part of the form definition are prefixed with `field_`, while
#'   columns from the "choices" part of the form definition are prefixed with
#'   `choice_`. Original column names from data returned by [scto_read()] are
#'   in the `submission_field_name` column, while corresponding values are in
#'   the `field_value` column. Data can be converted back to wide format as
#'   desired using [data.table::dcast()] or [tidyr::pivot_wider()].
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' responses = scto_get_form_responses(auth, 'my_form')
#' }
#'
#' @export
scto_get_form_responses = \(
  auth, form_id,
  id_cols = c(
    'KEY', 'formdef_version', 'CompletionDate', 'SubmissionDate',
    'starttime', 'endtime', 'review_status', 'review_quality'),
  exclude_cols = 'instanceID', ...) {

  . = ..field_cols = ..response_cols = `_form_version` = choice_value =
    field_name = field_type = field_value = list_name = submission_field_name =
    type = value = NULL

  meta = scto_get_form_metadata(auth, form_id)
  subs = scto_read(auth, form_id, ...)
  defs = scto_unnest_form_definitions(meta)

  assert_character(id_cols)
  assert_character(exclude_cols)

  fields = copy(defs$survey[[1L]])
  fields = fields[type %notin% c('begin group', 'end group')]
  setnames(fields, rename_cols)
  setnames(fields, c('name', '_row_num'), c('field_name', 'field_row_num'))
  setnames(
    fields, which(startsWith(colnames(fields), 'label')),
    \(x) paste0('field_', x))
  fields[, field_type := sub(' .+$', '', type)]
  fields[
    grepl('^select_(one|multiple) ', type), list_name := gsub('^.+ ', '', type)]

  choices = copy(defs$choices[[1L]])
  choices = choices[, .SD[1L], by = .(`_form_version`, list_name, value)]
  setnames(choices, rename_cols)
  setnames(choices, c('value', '_row_num'), c('choice_value', 'choice_row_num'))
  setnames(
    choices, which(startsWith(colnames(choices), 'label')),
    \(x) paste0('choice_', x))

  fields_choices = merge(
    fields[field_type %in% c('select_one', 'select_multiple')],
    choices, by = c('_form_version', 'list_name'), sort = FALSE,
    allow.cartesian = TRUE)

  fields_choices[, submission_field_name := fifelse(
    field_type == 'select_one', field_name,
    paste0(field_name, '_', choice_value))]

  measure_cols = setdiff(colnames(subs), c(id_cols, exclude_cols))
  responses = melt(
    subs, id.vars = id_cols, measure.vars = measure_cols,
    variable.name = 'submission_field_name', value.name = 'field_value')
  setnames(responses, rename_cols)

  by_x = c('formdef_version', 'submission_field_name')
  by_y = c('_form_version', 'submission_field_name')

  responses_fields = merge(
    responses, copy(fields)[, submission_field_name := field_name],
    by.x = by_x, by.y = by_y, all.x = TRUE, sort = FALSE)

  # field type would not be select_multiple
  x1 = merge(
    responses_fields[!is.na(field_name)],
    copy(choices)[, field_value := choice_value],
    by.x = c('formdef_version', 'list_name', 'field_value'),
    by.y = c('_form_version', 'list_name', 'field_value'),
    all.x = TRUE, sort = FALSE)

  # field type would be select_multiple, field value not NA
  withr::local_options(list(warn = -1)) # hide data.table warning
  response_cols = colnames(responses)
  x2 = merge(
    responses_fields[is.na(field_name) & !is.na(field_value), ..response_cols],
    fields_choices,
    by.x = by_x, by.y = by_y, all.x = TRUE, sort = FALSE)

  # field type would be select_multiple, field value NA
  field_cols = c(colnames(fields), 'submission_field_name')
  x3 = merge(
    responses_fields[is.na(field_name) & is.na(field_value), ..response_cols],
    unique(fields_choices[, ..field_cols]),
    by.x = by_x, by.y = by_y, all.x = TRUE, sort = FALSE)

  responses_fields_choices = rbind(x1, x2, x3, fill = TRUE) |>
    setcolorder(
      c('key', 'submission_field_name', 'field_value', 'formdef_version'))
  responses_fields_choices
}

#' Write data to a SurveyCTO server
#'
#' @description
#' `r lifecycle::badge('experimental')`
#'
#' This function updates an existing dataset using an unofficial API endpoint.
#' For this to work, the SurveyCTO user must have permissions "Can edit server
#' datasets" and "Can modify or delete server dataset data".
#'
#' @param auth [scto_auth()] object.
#' @param data `data.frame` to upload.
#' @param dataset_id String indicating id of existing dataset.
#' @param dataset_title String indicating title of dataset. Will replace the
#'   existing title, regardless of `append`.
#' @param append Logical indicating whether to append or replace the dataset.
#' @param fill Logical indicating whether to implicitly fill missing columns
#'   with `NA`, i.e., whether to allow a mismatch between columns of the
#'   existing dataset and columns of `data`. Only used if `append` is `TRUE`.
#'
#' @return A list with elements:
#'
#' * `data_old`: A `data.table` of the previous version of the dataset.
#' * `response`: An object of class [httr::response()] from the API call.
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' r = scto_write(auth, data, 'my_dataset', 'My Dataset')
#' }
#'
#' @export
scto_write = function(
    auth, data, dataset_id, dataset_title = dataset_id, append = FALSE,
    fill = FALSE) {
  assert_class(auth, 'scto_auth')
  assert_data_frame(data)
  assert_string(dataset_id)
  assert_string(dataset_title)
  assert_flag(append)

  # check that dataset exists
  data_old = scto_read(auth, dataset_id, drop_empty_cols = FALSE)
  if (attr(data_old, 'scto_type') != 'dataset') {
    scto_abort(
      'The id {.id {dataset_id}} corresponds to a form, not a dataset.')
  }

  if (append) {
    assert_flag(fill)
    if (!fill && !setequal(colnames(data), colnames(data_old))) {
      scto_abort(paste(
        'If `fill` is FALSE, column names of `data` must match',
        'those of the dataset {.dataset {dataset_id}}.'))
    }
  }

  path = withr::local_tempfile(fileext = '.csv')
  fwrite(data, path, logical01 = TRUE)

  # authentication
  upload_url = glue(
    'https://{auth$servername}.surveycto.com/',
    'datasets/{dataset_id}/upload?csrf_token={auth$csrf_token}')

  # data upload
  scto_bullets(c(v = 'Writing dataset {.dataset {dataset_id}}.'))

  dataset_upload_mode = if (append) 'append' else 'clear' # ignoring merge

  upload_res = POST(
    upload_url,
    body = list(
      dataset_exists = 1,
      dataset_id = dataset_id,
      dataset_title = dataset_title,
      dataset_upload_mode = dataset_upload_mode,
      dataset_type = 'SERVER',
      dataset_file = httr::upload_file(path)))

  r = list(data_old = data_old, response = upload_res)
  r
}

#' @import checkmate
#' @import cli
#' @import data.table
#' @importFrom glue glue
#' @importFrom httr GET POST content add_headers headers cookies set_cookies
NULL


assert_form_ids = function(auth, form_ids) {
  assert_character(
    form_ids, any.missing = FALSE, min.len = 1L, unique = TRUE, null.ok = TRUE)
  if (!is.null(form_ids)) return(form_ids)
  catalog = scto_catalog(auth)
  ids = catalog[catalog$type == 'form']$id
  ids
}


get_api_response = function(auth, request_url) {
  res = GET(request_url, add_headers('x-csrf-token' = auth$csrf_token))
  status = res$status_code
  content = rawToChar(res$content)
  if (status != 200L) {
    cli_alert_warning('Response content:\n{content}')
    scto_abort('Non-200 response: {status}')
  }
  content
}


#' Suppress or permit messages from rsurveycto
#'
#' By default, rsurveycto prints messages to the console. To suppress them, set
#' the `rsurveycto_quiet` option to `TRUE` or use this function.
#'
#' @param quiet A logical indicating whether to suppress messages, or `NULL`.
#'
#' @return If `quiet` is `NULL`, the current value of the `rsurveycto_quiet`
#'   option. Otherwise, the previous value of the `rsurveycto_quiet` option
#'   invisibly.
#'
#' @examples
#' options(rsurveycto_quiet = TRUE)
#' scto_quiet()
#' scto_quiet(FALSE)
#'
#' @export
scto_quiet = function(quiet = NULL) {
  assert_flag(quiet, null.ok = TRUE)
  quiet_old = getOption('rsurveycto_quiet')
  if (is.null(quiet)) return(quiet_old)
  options(rsurveycto_quiet = quiet)
  invisible(quiet_old)
}


scto_theme = function() {
  common = list(before = '"', after = '"')
  # Okabe-Ito colors
  list(
    span.server = c(color = '#E69F00', common), # orange
    span.id = c(color = '#D55E00', common), # vermillion
    span.dataset = c(color = '#56B4E9', common), # skyblue
    span.form = c(color = '#009E73', common), # bluishgreen
    span.version = c(color = '#F5C710', common), # amber
    span.filename = c(color = '#CC79A7', common)) # reddishpurple
}


scto_bullets = function(text, .envir = parent.frame()) {
  if (isTRUE(scto_quiet()) || identical(Sys.getenv('TESTTHAT'), 'true')) {
    return(invisible())
  }
  cli_div(theme = scto_theme())
  cli_bullets(text, .envir = .envir)
}


scto_abort = function(message, ..., .envir = parent.frame()) {
  call = rlang::caller_env()
  cli_div(theme = scto_theme())
  cli_abort(message = message, ..., .envir = .envir, call = call)
}


is_empty = function(x) {
  i = is.na(x)
  if (is.character(x)) i = i | x == ''
  all(i)
}


#' Drop empty columns from a data.table
#'
#' An empty column is one whose only values are `NA` or "".
#'
#' @param d `data.table`.
#'
#' @return `d` modified by reference, invisibly.
#'
#' @examples
#' library('data.table')
#' d = data.table(w = 3:4, x = c('', 'foo'), y = c(NA, NA), z = c(NA, ''))
#' drop_empties(d)
#'
#' @export
drop_empties = function(d) {
  assert_data_table(d)
  if (nrow(d) == 0) return(d)
  idx = sapply(d, is_empty)
  cols = colnames(d)[which(idx)]
  d[, c(cols) := NULL]
}

#' Fetch form metadata from a SurveyCTO server
#'
#' This function fetches metadata, including form definitions, for deployed and
#' previous versions of one or more forms.
#'
#' @param auth [scto_auth()] object.
#' @param form_ids Character vector indicating the form ids. `NULL` indicates
#'   all forms.
#' @param deployed_only Logical indicating whether to fetch metadata for all
#'   versions of each form, or only for the deployed version.
#' @param get_defs Logical indicating whether to fetch form definitions.
#' @param def_dir String indicating directory in which to save the form
#'   definitions as Excel files.
#'
#' @return A `data.table` with one row per form (and per version, if
#'   `deployed_only` is `FALSE`). Definitions are returned as nested
#'   `data.table`s, which can be unnested using
#'   [scto_unnest_form_definitions()].
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' form_metadata = scto_get_form_metadata(auth, 'my_form')
#' form_defs = scto_unnest_form_definitions(form_metadata)
#' }
#'
#' @export
scto_get_form_metadata = function(
    auth, form_ids = NULL, deployed_only = FALSE, get_defs = TRUE,
    def_dir = NULL) {
  assert_class(auth, 'scto_auth')
  form_ids = assert_form_ids(auth, form_ids)
  assert_flag(deployed_only)
  assert_flag(get_defs)
  if (get_defs) {
    assert_string(def_dir, null.ok = TRUE)
    if (!is.null(def_dir)) {
      assert_path_for_output(def_dir, overwrite = TRUE)
    }
  }

  r = list()
  for (i in seq_along(form_ids)) {
    id = form_ids[i]
    cot = tryCatch(
      get_form_meta(auth, id, deployed_only, get_defs, def_dir), error = \(e) e)
    if (inherits(cot, 'error')) scto_abort('Form {.form {id}} was not found.')
    r[[i]] = cot
  }

  d = rbindlist(r, use.names = TRUE, fill = TRUE)
  d
}


get_form_meta = function(auth, id, deployed_only, get_defs, def_dir) {
  unix_ms = as.numeric(Sys.time()) * 1000
  request_url = glue(
    'https://{auth$servername}.surveycto.com/forms/{id}/files?t={unix_ms}')

  scto_bullets(c(v = 'Reading metadata for form {.form {id}}.'))
  content = get_api_response(auth, request_url)

  r = jsonlite::fromJSON(content)
  d = as.data.table(r$deployedGroupFiles$definitionFile)
  if (nrow(d) == 0L) return(d)

  if (!deployed_only) d = rbind(d, as.data.table(r$previousDefinitionFiles))
  set(d, j = 'id', value = NULL)
  set(d, j = 'is_deployed', value = c(TRUE, rep_len(FALSE, nrow(d) - 1L)))
  setnames(
    d, c('formVersion', 'downloadLink', 'dateStr'),
    c('form_version', 'download_link', 'date_str'))
  set(d, j = 'form_id', value = id)
  setcolorder(d, 'form_id')

  if (isFALSE(get_defs)) return(d)

  defs = lapply(seq_len(nrow(d)), \(i) {
    get_form_def_excel(auth, d$download_link[i], d$form_version[i], id, def_dir)
  })
  for (j in names(defs[[1L]])) set(d, j = j, value = lapply(defs, \(x) x[[j]]))
  d
}


get_form_def_excel = function(auth, url, ver, id, def_dir) {
  value = NULL
  path = withr::local_tempfile()
  scto_bullets(
    c(v = 'Fetching definition for form version {.version {ver}}.'))
  curl::curl_download(url, path, handle = auth$handle)

  sheets = c('survey', 'choices', 'settings')
  f = \(...) vctrs::vec_as_names(..., repair = 'unique_quiet')

  r = sapply(sheets, \(sheet) {
    d = readxl::read_excel(path, sheet, col_types = 'text', .name_repair = f)
    setDT(d)
    # with or without col_types = 'text', was sometimes getting 1.0 instead of 1
    if (sheet == 'choices') {
      if ('value' %in% colnames(d)) {
        withr::local_options(list(warn = -1))
        d[, value := fifelse(
          !is.na(as.integer(value)) & endsWith(value, '.0'),
          as.character(as.integer(value)), value)]
      } else {
        d[, value := NA_character_]
      }
    }
    # d = readxl::read_excel(path, sheet, guess_max = 1e4, .name_repair = f)
    # setDT(d)
    # cols = colnames(d)[!sapply(d, is.character)]
    # for (j in cols) set(d, j = j, value = as.character(d[[j]]))
    d
  })

  if (!is.null(def_dir)) {
    ext = if (grepl('\\?file=.+\\.xlsx&', tolower(url))) 'xlsx' else 'xls'
    # file name conflicts are possible, but hopefully pathological
    file.rename(path, file.path(def_dir, glue('{id}__{ver}.{ext}')))
  }

  r
}


#' Unnest previously fetched form definitions
#'
#' This function unnests form definitions, e.g., from multiple versions of a
#' form, which can make it easier to map values to labels in a later step.
#'
#' @param form_metadata `data.table` returned by [scto_get_form_metadata()].
#' @param by_form_id Logical indicating whether to unnest definitions of
#'   multiple versions of a given form (default), or to unnest definitions of
#'   all forms together.
#'
#' @return If `by_form_id` is `TRUE`, a `data.table` of `data.table`s for the
#'   survey, choices, and settings components of the form definitions. Otherwise
#'   a list of `data.table`s.
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' form_metadata = scto_get_form_metadata(auth, 'my_form')
#' form_defs = scto_unnest_form_definitions(form_metadata)
#' }
#'
#' @export
scto_unnest_form_definitions = function(form_metadata, by_form_id = TRUE) {
  assert_data_table(form_metadata)
  assert_flag(by_form_id)
  def_cols = c('survey', 'choices', 'settings')
  assert_names(
    colnames(form_metadata),
    must.include = c('form_id', 'form_version', def_cols))

  . = form_id = form_version = NULL # nolint

  form_versions = form_metadata[
    , .(.id = seq_len(.N), `_form_id` = form_id,
        `_form_version` = form_version)]

  if (isTRUE(by_form_id)) {
    r = lapply(unique(form_metadata$form_id), \(form_id_now) {
      form_metadata_now = form_metadata[form_id == form_id_now]
      unnest_form_defs(
        form_metadata_now, form_versions, def_cols, '_form_version')
    })
    names(r) = unique(form_metadata$form_id)
    r = as.data.table(do.call(rbind, r), keep.rownames = 'form_id')

  } else {
    r = unnest_form_defs(
      form_metadata, form_versions, def_cols, c('_form_id', '_form_version'))
  }
  r
}


unnest_form_defs = function(form_metadata, form_versions, def_cols, form_cols) {
  .id = `_row_num` = NULL # nolint
  r = list()
  for (j in def_cols) {
    r[[j]] = rbindlist(
      form_metadata[[j]], use.names = TRUE, fill = TRUE, idcol = TRUE) |>
      merge(form_versions[, c('.id', form_cols), with = FALSE],
            by = '.id', sort = FALSE)
    r[[j]][, `_row_num` := seq_len(.N), by = .id]
    r[[j]][, .id := NULL]
    setcolorder(r[[j]], form_cols)[]
  }
  r
}

if (identical(Sys.getenv('NOT_CRAN'), 'true')) { # !on_cran()
  auth = scto_auth(auth_file = auth_file)
  id = 'hh_listing_example_1'
  nms = c(
    'form_id', 'form_version', 'is_deployed', 'survey', 'choices', 'settings')
}

test_that('scto_get_form_metadata one', {
  skip_on_cran()
  meta = scto_get_form_metadata(auth, id)
  expect_data_table(meta)
  expect_names(names(meta), must.include = nms)
})

test_that('scto_get_form_metadata all deployed', {
  skip_on_cran()
  meta = scto_get_form_metadata(auth, deployed_only = TRUE)
  expect_data_table(meta)
  expect_names(names(meta), must.include = nms)
  expect_setequal(meta$is_deployed, 1L)
})

test_that('scto_get_form_metadata get_defs false', {
  skip_on_cran()
  meta = scto_get_form_metadata(auth, id, get_defs = FALSE)
  expect_data_table(meta)
  expect_names(names(meta), disjunct.from = nms[4:6])
})

test_that('scto_get_form_metadata not ok', {
  skip_on_cran()
  expect_error(scto_get_form_metadata(auth, 'flux_capacitors'))
})

test_that('scto_unnest_form_definitions', {
  skip_on_cran()
  meta = scto_get_form_metadata(auth, id)
  defs = scto_unnest_form_definitions(meta)
  expect_data_table(defs, nrows = 1L)
  expect_names(names(defs), permutation.of = nms[c(1, 4:6)])
})

test_that('scto_unnest_form_definitions by_form_id false', {
  skip_on_cran()
  meta = scto_get_form_metadata(auth, id)
  defs = scto_unnest_form_definitions(meta, by_form_id = FALSE)
  expect_list(defs)
  expect_names(names(defs), permutation.of = nms[4:6])
})

#' Fetch file attachments from a SurveyCTO server
#'
#' This function can download encrypted and unencrypted files attached to forms.
#'
#' @param auth [scto_auth()] object.
#' @param urls Character vector of API URLs for file attachments. Will typically
#'   be derived from a column of a `data.table` returned by [scto_read()]. Can
#'   contain missing values.
#' @param output_dir String indicating path to directory in which to save files.
#' @param private_key String indicating path to private key file. Only needs to
#'   be non-`NULL` to decrypt encrypted file attachments.
#' @param overwrite Logical indicating whether to overwrite existing files.
#'
#' @return A character vector of file names of the same length as `urls`, with
#'   `NA` for missing or invalid URLs.
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' form_data = scto_read(auth, 'my_form', 'form')
#' filenames = scto_get_attachments(auth, form_data[['my_attachment']], '.')
#' }
#'
#' @export
scto_get_attachments = function(
    auth, urls, output_dir, private_key = NULL, overwrite = TRUE) {

  assert_class(auth, 'scto_auth')
  assert_character(urls)
  assert_string(output_dir)
  assert_string(private_key, null.ok = TRUE)
  if (!is.null(private_key)) assert_file_exists(private_key)
  assert_flag(overwrite)

  pat = paste0(
    '^https://[a-z0-9_-]+\\.surveycto\\.com/api/v2/forms/[a-z0-9_-]+/',
    'submissions/uuid:[a-z0-9-]+/attachments/[a-zA-Z0-9_.-]+$')

  r = rep(NA_character_, length(urls))
  idx = grepl(pat, urls)
  if (!any(idx)) return(r)

  urls = urls[idx]
  filenames = basename(urls) # depends on SurveyCTO making filenames unique

  # devtools said these longer filenames were non-portable
  # x = strsplit(urls, '/')
  # filenames = paste0(sapply(x, `[[`, 9L), '__', sapply(x, `[[`, 11L))
  # filenames = sub('^uuid:', '', filenames)

  coll = makeAssertCollection()
  for (filename in filenames) {
    assert_path_for_output(
      file.path(output_dir, filename), overwrite = overwrite, add = coll)
  }
  reportAssertions(coll)

  if (!dir.exists(output_dir)) dir.create(output_dir, recursive = TRUE)

  for (i in seq_along(urls)) {
    scto_bullets(c(v = 'Downloading {.filename {filenames[i]}}.'))
    path = file.path(output_dir, filenames[i])
    if (is.null(private_key)) {
      res = curl::curl_download(urls[i], path, handle = auth$handle)
    } else {
      res = POST(
        urls[i], body = list(private_key = httr::upload_file(private_key)))
      writeBin(res$content, path)
    }
  }

  r[idx] = filenames
  r
}

## for cran submission:
# with(list(f = 'tests/testthat/scto_auth.txt'),
#      devtools::check_rhub(
#        env_vars = c(SCTO_AUTH = readChar(f, file.info(f)$size))))

library('data.table')

output_dir = withr::local_tempdir(.local_envir = teardown_env())
withr::local_options(warn = -1L, .local_envir = teardown_env())

if (Sys.getenv('SCTO_AUTH') == '') {
  auth_file = 'scto_auth.txt'
} else {
  auth_file = withr::local_tempfile(.local_envir = teardown_env())
  writeLines(Sys.getenv('SCTO_AUTH'), auth_file)
}

auth_args = readLines(auth_file)

if (Sys.getenv('SCTO_PRIVATE_KEY') == '') {
  private_key = 'rsurveycto-private-key.pem'
} else {
  private_key = withr::local_tempfile(.local_envir = teardown_env())
  writeLines(Sys.getenv('SCTO_PRIVATE_KEY'), private_key)
}

#' Read metadata from a SurveyCTO server
#'
#' These functions read metadata from a SurveyCTO server.
#'
#' @param auth [scto_auth()] object.
#'
#' @return `scto_meta()` returns a nested list of metadata related to forms,
#'   datasets, groups, and publishing information. `scto_catalog()` returns a
#'   `data.table` with one row per form or dataset.
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' metadata = scto_meta(auth)
#' catalog = scto_catalog(auth)
#' }
#'
#' @export
scto_meta = function(auth) {
  assert_class(auth, 'scto_auth')
  url = glue(
    'https://{auth$servername}.surveycto.com/console/forms-groups-datasets/get')

  res = GET(url, set_cookies(JSESSIONID = auth$session_id),
            add_headers('x-csrf-token' = auth$csrf_token))

  if (res$status_code != 200L) {
    scto_abort(
      'Invalid username or password for server {.server {auth$servername}}.')
  }

  scto_bullets(
    c(v = 'Reading metadata for server {.server {auth$servername}}.'))
  m = content(res, 'parsed')
  m
}


#' @rdname scto_meta
#' @export
scto_catalog = function(auth) {
  # surveycto enforces uniqueness of ids across forms and datasets
  creationDate = last_version_created_at = last_incoming_data_at = # nolint
    lastIncomingDataDate = form_version = type = NULL # nolint
  m = scto_meta(auth)
  common_cols = c('id', 'title', 'version', 'groupId')

  if (length(m$datasets) > 0L) {
    dataset_cols = c(common_cols, 'casesDataset', 'discriminator', 'rows')
    datasets = rbindlist(
      lapply(m$datasets, \(x) x[dataset_cols]), use.names = TRUE, fill = TRUE)
    setnames(datasets, c('casesDataset', 'version', 'rows', 'groupId'),
             c('is_cases_dataset', 'dataset_version', 'num_rows', 'group_id'))
  } else {
    datasets = data.table()
  }

  if (length(m$forms) > 0L) {
    sub_idx = grepl('SubmissionCount$', names(m$forms[[1L]]))
    sub_cols = names(m$forms[[1L]])[sub_idx]
    form_cols = c(
      common_cols, 'creationDate', 'lastIncomingDataDate', 'encrypted',
      'deployed', 'reviewWorkflowEnabled', sub_cols)

    forms = rbindlist(
      lapply(m$forms, \(x) x[form_cols]), use.names = TRUE, fill = TRUE)
    forms[, last_version_created_at := as.POSIXct(
      creationDate / 1000, tz = 'UTC')]
    forms[, last_incoming_data_at := as.POSIXct(
      lastIncomingDataDate / 1000, tz = 'UTC')]
    forms[, `:=`(creationDate = NULL, lastIncomingDataDate = NULL)]

    old = c(
      'version', 'groupId', 'encrypted', 'deployed', 'reviewWorkflowEnabled',
      sub_cols)
    new = c(
      'form_version', 'group_id', 'is_encrypted', 'is_deployed',
      'is_review_workflow_enabled',
      paste0('num_submissions_', gsub('SubmissionCount', '', sub_cols)))
    setnames(forms, old, new)
  } else {
    forms = data.table()
  }

  d = rbind(forms, datasets, use.names = TRUE, fill = TRUE)
  d[, type := fifelse(is.na(form_version), 'dataset', 'form')]

  g = rbindlist(lapply(m$groups, \(x) x[c('id', 'title')]))
  setnames(g, c('group_id', 'group_title'))
  d = merge(d, g, by = 'group_id', sort = FALSE)
  neworder = c(
    'id', 'title', 'type', 'form_version', 'dataset_version',
    'last_version_created_at', 'last_incoming_data_at', 'is_encrypted',
    'is_deployed', 'is_review_workflow_enabled', 'is_cases_dataset',
    'discriminator', 'group_id', 'group_title')
  setcolorder(d, neworder, skip_absent = TRUE)
  setkey(d)
}

if (identical(Sys.getenv('NOT_CRAN'), 'true')) { # !on_cran()
  auth = scto_auth(auth_file = auth_file)
}

test_that('scto_meta ok', {
  skip_on_cran()
  meta = scto_meta(auth)

  nms = c('canAddObjectsIntoRoot', 'groups', 'datasets',
          'serverDatasetPublishingInfo', 'forms')

  expect_type(meta, 'list')
  expect_named(meta, nms)
})

test_that('scto_catalog ok', {
  skip_on_cran()
  catalog = scto_catalog(auth)
  expect_data_table(catalog)
})

#' Fetch deployed form definitions from a SurveyCTO server
#'
#' @description
#' `r lifecycle::badge('superseded')`
#'
#' This function fetches definitions for currently deployed forms. It has been
#' superseded in favor of [scto_get_form_metadata()], which fetches metadata,
#' including defintions, for deployed and previous versions of forms.
#'
#' @param auth [scto_auth()] object.
#' @param form_ids Character vector indicating the form ids. `NULL` indicates
#'   all forms.
#' @param simplify Logical indicating whether to return the definition for one
#'   form as a simple list instead of a named, nested list.
#'
#' @return If `simplify` is `TRUE` and getting one form definition, a list.
#'   Otherwise a named list of lists containing the definition for each form.
#'
#' @examples
#' \dontrun{
#' auth = scto_auth('scto_auth.txt')
#' form_def = scto_get_form_definitions(auth, 'my_form')
#' form_defs = scto_get_form_definitions(auth)
#' }
#'
#' @export
scto_get_form_definitions = function(auth, form_ids = NULL, simplify = TRUE) {
  assert_character(form_ids, any.missing = FALSE, unique = TRUE, null.ok = TRUE)

  assert_class(auth, 'scto_auth')
  form_ids = assert_form_ids(auth, form_ids)
  assert_flag(simplify)

  # works even if no forms
  r = list()
  for (i in seq_along(form_ids)) {
    id = form_ids[i]
    cot = tryCatch(get_form_def(auth, id), error = \(e) e)
    if (inherits(cot, 'error')) scto_abort('Form {.form {id}} was not found.')
    r[[i]] = cot
  }
  names(r) = form_ids
  if (length(r) == 1L && simplify) r = r[[1L]]
  r
}


get_form_def = function(auth, id) {
  request_url = glue(
    'https://{auth$servername}.surveycto.com/forms/{id}/design')

  scto_bullets(c(v = 'Fetching definition for form {.form {id}}.'))
  content = get_api_response(auth, request_url)

  r = jsonlite::fromJSON(content)
  idx = which(sapply(r, \(x) inherits(x, 'matrix')))
  for (i in idx) {
    cols = r[[i]][1L, ]
    r[[i]] = as.data.table(r[[i]][-1L, , drop = FALSE])
    setnames(r[[i]], cols)
    for (j in colnames(r[[i]])) {
      set(r[[i]], i = which(r[[i]][[j]] == ''), j = j, value = NA_character_)
    }
  }
  names(r) = sub('RowsAndColumns$', '', names(r))
  r
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_write.R
\name{scto_write}
\alias{scto_write}
\title{Write data to a SurveyCTO server}
\usage{
scto_write(
  auth,
  data,
  dataset_id,
  dataset_title = dataset_id,
  append = FALSE,
  fill = FALSE
)
}
\arguments{
\item{auth}{\code{\link[=scto_auth]{scto_auth()}} object.}

\item{data}{\code{data.frame} to upload.}

\item{dataset_id}{String indicating id of existing dataset.}

\item{dataset_title}{String indicating title of dataset. Will replace the
existing title, regardless of \code{append}.}

\item{append}{Logical indicating whether to append or replace the dataset.}

\item{fill}{Logical indicating whether to implicitly fill missing columns
with \code{NA}, i.e., whether to allow a mismatch between columns of the
existing dataset and columns of \code{data}. Only used if \code{append} is \code{TRUE}.}
}
\value{
A list with elements:
\itemize{
\item \code{data_old}: A \code{data.table} of the previous version of the dataset.
\item \code{response}: An object of class \code{\link[httr:response]{httr::response()}} from the API call.
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function updates an existing dataset using an unofficial API endpoint.
For this to work, the SurveyCTO user must have permissions "Can edit server
datasets" and "Can modify or delete server dataset data".
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
r = scto_write(auth, data, 'my_dataset', 'My Dataset')
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_auth.R
\name{scto_auth}
\alias{scto_auth}
\title{Authenticate with a SurveyCTO server}
\usage{
scto_auth(
  auth_file = NULL,
  servername = NULL,
  username = NULL,
  password = NULL,
  validate = TRUE
)
}
\arguments{
\item{auth_file}{String indicating path to a text file containing the
server name on the first line, username on the second, and password on the
third. Other arguments are only used if \code{auth_file} is \code{NULL}.}

\item{servername}{String indicating name of the SurveyCTO server.}

\item{username}{String indicating username for the SurveyCTO account.}

\item{password}{String indicating password for the SurveyCTO account.}

\item{validate}{Logical indicating whether to validate credentials by calling
\code{\link[=scto_meta]{scto_meta()}}. Should only be set to \code{FALSE} for debugging.}
}
\value{
\code{scto_auth} object for an authenticated SurveyCTO session.
}
\description{
SurveyCTO's API supports basic authentication using a username and password.
Make sure the user is assigned a role with permission to download data
("data manager" or greater) and "Allow server API access" is enabled.
}
\examples{
\dontrun{
# preferred approach
auth = scto_auth('scto_auth.txt')

# alternate approach
auth = scto_auth('my_server', 'my_user', 'my_pw', auth_file = NULL)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_get_attachments.R
\name{scto_get_attachments}
\alias{scto_get_attachments}
\title{Fetch file attachments from a SurveyCTO server}
\usage{
scto_get_attachments(
  auth,
  urls,
  output_dir,
  private_key = NULL,
  overwrite = TRUE
)
}
\arguments{
\item{auth}{\code{\link[=scto_auth]{scto_auth()}} object.}

\item{urls}{Character vector of API URLs for file attachments. Will typically
be derived from a column of a \code{data.table} returned by \code{\link[=scto_read]{scto_read()}}. Can
contain missing values.}

\item{output_dir}{String indicating path to directory in which to save files.}

\item{private_key}{String indicating path to private key file. Only needs to
be non-\code{NULL} to decrypt encrypted file attachments.}

\item{overwrite}{Logical indicating whether to overwrite existing files.}
}
\value{
A character vector of file names of the same length as \code{urls}, with
\code{NA} for missing or invalid URLs.
}
\description{
This function can download encrypted and unencrypted files attached to forms.
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
form_data = scto_read(auth, 'my_form', 'form')
filenames = scto_get_attachments(auth, form_data[['my_attachment']], '.')
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_read.R
\name{scto_read}
\alias{scto_read}
\title{Read data from a SurveyCTO server}
\usage{
scto_read(
  auth,
  ids = NULL,
  start_date = as.POSIXct("1900-01-01", tz = "UTC"),
  review_status = "approved",
  private_key = NULL,
  drop_empty_cols = FALSE,
  convert_datetime = c("CompletionDate", "SubmissionDate", "starttime", "endtime"),
  datetime_format = "\%b \%e, \%Y \%I:\%M:\%S \%p",
  simplify = TRUE
)
}
\arguments{
\item{auth}{\code{\link[=scto_auth]{scto_auth()}} object.}

\item{ids}{Character vector indicating ids of the datasets and/or forms.
\code{NULL} indicates all datasets and forms.}

\item{start_date}{Date-time or something coercible to a date-time
indicating the earliest date-time (UTC timezone) for which to fetch data.
Only used for forms. Use with caution, because fields that are deleted
prior to \code{start_date} will not show up, even if submissions prior to
\code{start_date} have data for those fields.}

\item{review_status}{String or character vector indicating which submissions
to fetch. Possible values are "approved", "pending", "rejected", or any
combination of the three. Only used for forms.}

\item{private_key}{String indicating path to private key file. Only needs to
be non-\code{NULL} to read encrypted form data.}

\item{drop_empty_cols}{Logical indicating whether to drop columns that
contain only \code{NA} or only an empty string.}

\item{convert_datetime}{Character vector of column names in the data for
which to convert strings to datetimes (POSIXct). Use \code{NULL} to not convert
any columns to datetimes.}

\item{datetime_format}{String indicating format of datetimes from SurveyCTO.
See \code{\link[=strptime]{strptime()}}.}

\item{simplify}{Logical indicating whether to return only a \code{data.table}
instead of a list of \code{data.table}s if reading one form or dataset.}
}
\value{
If \code{simplify} is \code{TRUE} and reading one form or dataset, a
\code{data.table}. Otherwise a named list of \code{data.table}s, one for each form
and dataset, along with a \code{data.table} named ".catalog" from
\code{scto_catalog()}.
}
\description{
This function can read datasets and forms.
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
form_data = scto_read(auth, 'my_form')
all_data = scto_read(auth)
}

}
\seealso{
\code{\link[=scto_get_form_responses]{scto_get_form_responses()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_get_form_metadata.R
\name{scto_unnest_form_definitions}
\alias{scto_unnest_form_definitions}
\title{Unnest previously fetched form definitions}
\usage{
scto_unnest_form_definitions(form_metadata, by_form_id = TRUE)
}
\arguments{
\item{form_metadata}{\code{data.table} returned by \code{\link[=scto_get_form_metadata]{scto_get_form_metadata()}}.}

\item{by_form_id}{Logical indicating whether to unnest definitions of
multiple versions of a given form (default), or to unnest definitions of
all forms together.}
}
\value{
If \code{by_form_id} is \code{TRUE}, a \code{data.table} of \code{data.table}s for the
survey, choices, and settings components of the form definitions. Otherwise
a list of \code{data.table}s.
}
\description{
This function unnests form definitions, e.g., from multiple versions of a
form, which can make it easier to map values to labels in a later step.
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
form_metadata = scto_get_form_metadata(auth, 'my_form')
form_defs = scto_unnest_form_definitions(form_metadata)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsurveycto-package.R
\docType{package}
\name{rsurveycto-package}
\alias{rsurveycto}
\alias{rsurveycto-package}
\title{rsurveycto: Interact with Data on 'SurveyCTO'}
\description{
'SurveyCTO' is a platform for mobile data collection in offline settings. The 'rsurveycto' R package uses the 'SurveyCTO' REST API \url{https://docs.surveycto.com/05-exporting-and-publishing-data/05-api-access/01.api-access.html} to read datasets and forms from a 'SurveyCTO' server into R as 'data.table's and to download file attachments. The package also has limited support to write datasets to a server.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://agency-fund.github.io/rsurveycto/}
  \item \url{https://github.com/agency-fund/rsurveycto}
}

}
\author{
\strong{Maintainer}: Jake Hughey \email{jake@agency.fund}

Authors:
\itemize{
  \item Robert On \email{robert@agency.fund}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_get_form_responses.R
\name{scto_get_form_responses}
\alias{scto_get_form_responses}
\title{Read form data from a SurveyCTO server and convert to long format}
\usage{
scto_get_form_responses(
  auth,
  form_id,
  id_cols = c("KEY", "formdef_version", "CompletionDate", "SubmissionDate", "starttime",
    "endtime", "review_status", "review_quality"),
  exclude_cols = "instanceID",
  ...
)
}
\arguments{
\item{auth}{\code{\link[=scto_auth]{scto_auth()}} object.}

\item{form_id}{String indicating id of the form.}

\item{id_cols}{Character vector of data columns to keep in the result, will
be passed as \code{id.vars} to \code{\link[data.table:melt.data.table]{data.table::melt()}}.}

\item{exclude_cols}{Character vector of data columns to exclude from the
result.}

\item{...}{Additional arguments passed to \code{\link[=scto_read]{scto_read()}}.}
}
\value{
A \code{data.table} with one row per response, i.e., one row per form
submission per field. If necessary to avoid ambiguity, columns from the
"survey" part of the form definition are prefixed with \code{field_}, while
columns from the "choices" part of the form definition are prefixed with
\code{choice_}. Original column names from data returned by \code{\link[=scto_read]{scto_read()}} are
in the \code{submission_field_name} column, while corresponding values are in
the \code{field_value} column. Data can be converted back to wide format as
desired using \code{\link[data.table:dcast.data.table]{data.table::dcast()}} or \code{\link[tidyr:pivot_wider]{tidyr::pivot_wider()}}.
}
\description{
This function converts data from \code{\link[=scto_read]{scto_read()}} to long format, then joins
the data with the "survey" and "choices" components of the form metadata.
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
responses = scto_get_form_responses(auth, 'my_form')
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_meta.R
\name{scto_meta}
\alias{scto_meta}
\alias{scto_catalog}
\title{Read metadata from a SurveyCTO server}
\usage{
scto_meta(auth)

scto_catalog(auth)
}
\arguments{
\item{auth}{\code{\link[=scto_auth]{scto_auth()}} object.}
}
\value{
\code{scto_meta()} returns a nested list of metadata related to forms,
datasets, groups, and publishing information. \code{scto_catalog()} returns a
\code{data.table} with one row per form or dataset.
}
\description{
These functions read metadata from a SurveyCTO server.
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
metadata = scto_meta(auth)
catalog = scto_catalog(auth)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{scto_quiet}
\alias{scto_quiet}
\title{Suppress or permit messages from rsurveycto}
\usage{
scto_quiet(quiet = NULL)
}
\arguments{
\item{quiet}{A logical indicating whether to suppress messages, or \code{NULL}.}
}
\value{
If \code{quiet} is \code{NULL}, the current value of the \code{rsurveycto_quiet}
option. Otherwise, the previous value of the \code{rsurveycto_quiet} option
invisibly.
}
\description{
By default, rsurveycto prints messages to the console. To suppress them, set
the \code{rsurveycto_quiet} option to \code{TRUE} or use this function.
}
\examples{
options(rsurveycto_quiet = TRUE)
scto_quiet()
scto_quiet(FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{drop_empties}
\alias{drop_empties}
\title{Drop empty columns from a data.table}
\usage{
drop_empties(d)
}
\arguments{
\item{d}{\code{data.table}.}
}
\value{
\code{d} modified by reference, invisibly.
}
\description{
An empty column is one whose only values are \code{NA} or "".
}
\examples{
library('data.table')
d = data.table(w = 3:4, x = c('', 'foo'), y = c(NA, NA), z = c(NA, ''))
drop_empties(d)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_get_form_definitions.R
\name{scto_get_form_definitions}
\alias{scto_get_form_definitions}
\title{Fetch deployed form definitions from a SurveyCTO server}
\usage{
scto_get_form_definitions(auth, form_ids = NULL, simplify = TRUE)
}
\arguments{
\item{auth}{\code{\link[=scto_auth]{scto_auth()}} object.}

\item{form_ids}{Character vector indicating the form ids. \code{NULL} indicates
all forms.}

\item{simplify}{Logical indicating whether to return the definition for one
form as a simple list instead of a named, nested list.}
}
\value{
If \code{simplify} is \code{TRUE} and getting one form definition, a list.
Otherwise a named list of lists containing the definition for each form.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

This function fetches definitions for currently deployed forms. It has been
superseded in favor of \code{\link[=scto_get_form_metadata]{scto_get_form_metadata()}}, which fetches metadata,
including defintions, for deployed and previous versions of forms.
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
form_def = scto_get_form_definitions(auth, 'my_form')
form_defs = scto_get_form_definitions(auth)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scto_get_form_metadata.R
\name{scto_get_form_metadata}
\alias{scto_get_form_metadata}
\title{Fetch form metadata from a SurveyCTO server}
\usage{
scto_get_form_metadata(
  auth,
  form_ids = NULL,
  deployed_only = FALSE,
  get_defs = TRUE,
  def_dir = NULL
)
}
\arguments{
\item{auth}{\code{\link[=scto_auth]{scto_auth()}} object.}

\item{form_ids}{Character vector indicating the form ids. \code{NULL} indicates
all forms.}

\item{deployed_only}{Logical indicating whether to fetch metadata for all
versions of each form, or only for the deployed version.}

\item{get_defs}{Logical indicating whether to fetch form definitions.}

\item{def_dir}{String indicating directory in which to save the form
definitions as Excel files.}
}
\value{
A \code{data.table} with one row per form (and per version, if
\code{deployed_only} is \code{FALSE}). Definitions are returned as nested
\code{data.table}s, which can be unnested using
\code{\link[=scto_unnest_form_definitions]{scto_unnest_form_definitions()}}.
}
\description{
This function fetches metadata, including form definitions, for deployed and
previous versions of one or more forms.
}
\examples{
\dontrun{
auth = scto_auth('scto_auth.txt')
form_metadata = scto_get_form_metadata(auth, 'my_form')
form_defs = scto_unnest_form_definitions(form_metadata)
}

}
